import os
from pathlib import Path
from tqdm import tqdm
import geopandas as gpd

from ..auth import with_auth
from .retrieve import retrieve_dataset
from ..repos import FilesAPIRepo
from ..files.metadata import Metadata

@with_auth
def stage_dataset(
    dataset_name,
    version=None,
    path=None,
    logger=print,
    assets=False,
    force=False,
    verbose=False,
    user=None,
    file=None,
):
    dataset = retrieve_dataset(dataset_name, user)
    if version is None:
        version = sorted([v['version_id'] for v in dataset["versions"]])[-1]
    else:
        assert version in [
            v["version_id"] for v in dataset["versions"]
        ], f"Version {version} not found"
    download_base_path = os.getenv(
        "EOTDL_DOWNLOAD_PATH", str(Path.home()) + "/.cache/eotdl/datasets"
    )
    if path is None:
        download_path = download_base_path + "/" + dataset_name 
    else:
        download_path = path + "/" + dataset_name 
    # check if dataset already exists
    if os.path.exists(download_path) and not force:
        os.makedirs(download_path, exist_ok=True)
        # raise Exception(
        #     f"Dataset `{dataset['name']} v{str(version)}` already exists at {download_path}. To force download, use force=True or -f in the CLI."
        # )
        raise Exception(
            f"Dataset `{dataset['name']}` already exists at {download_path}. To force download, use force=True or -f in the CLI."
        )
    # stage metadata
    repo = FilesAPIRepo()
    catalog_path = repo.stage_file(dataset["id"], f"catalog.v{version}.parquet", user, download_path)
    # stage README.md
    metadata = Metadata(**dataset['metadata'], name=dataset['name'])
    metadata.save_metadata(download_path)
    # stage assets
    if assets:
        gdf = gpd.read_parquet(catalog_path)
        for _, row in tqdm(gdf.iterrows(), total=len(gdf), desc="Staging assets"):
            for k, v in row["assets"].items():
                stage_dataset_file(v["href"], download_path)
    return download_path


@with_auth
def stage_dataset_file(file_url, path, user):
    repo = FilesAPIRepo()
    return repo.stage_file_url(file_url, path, user)



from glob import glob
import json
from sklearn.model_selection import train_test_split

from .ds import MultiTaskDataset
from ..classification.dm import SCANEOClassificationDataModule, get_classification_labels

class SCANEOMultiTaskDataModule(SCANEOClassificationDataModule):
    def __init__(self, 
            path, 
            classes, 
            batch_size=16, 
            num_workers=4, 
            pin_memory=True, 
            val_split=0.2, 
            train_trans=None, 
            val_trans=None):
        super().__init__(
            path=path,
            classes=classes,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            val_split=val_split,
            train_trans=train_trans,
            val_trans=val_trans
        )
        self.classification_classes = classes
        self.num_cls_classes = len(classes)
        with open(f'{path}/spai.json', 'r') as f:
            self.spai_labels = json.load(f)
        self.classes = [label['name'] for label in self.spai_labels['labels']]
        self.num_seg_classes = len(self.classes) + 1
        self.colors = {label['name']: label['color'] for label in self.spai_labels['labels']}

    def setup(self, stage=None):
        masks = sorted(glob(f'{self.path}/*_mask.tif'))
        images = [i.replace('_mask.tif', '.tif') for i in masks]
        labels = [i.replace('tif', 'geojson') for i in images]
        images, classification_labels = get_classification_labels(images, labels, self.classification_classes)
        train_image, val_image, train_label, val_label = train_test_split(images, classification_labels, test_size=self.val_split, random_state=42)
        train_mask = [i.replace('.tif', '_mask.tif') for i in train_image]
        val_mask = [i.replace('.tif', '_mask.tif') for i in val_image]
        print("Training on", len(train_image), "images")
        print("Validating on", len(val_image), "images")
        self.train_ds = MultiTaskDataset(train_image, train_mask, train_label, trans=self.train_trans, num_classes=self.num_seg_classes)
        self.val_ds = MultiTaskDataset(val_image, val_mask, val_label, trans=self.val_trans, num_classes=self.num_seg_classes)


import pandas as pd
from typing import List

def compute_temporal_extent(start_date: str, nb_months: int) -> List[str]:
    """Compute temporal extent based on a start date and duration in months.

    Args:
        start_date (str): Start date in 'YYYY-MM-DD' format.
        nb_months (int): Number of months from the start date.

    Returns:
        List[str]: Temporal extent as [start_date, end_date].
    """
    start = pd.to_datetime(start_date)
    end = start + pd.DateOffset(months=nb_months)
    return [start.strftime('%Y-%m-%d'), end.strftime('%Y-%m-%d')]


"""
Sentinel-Hub data access module.
"""

from .client import SHClient
from .parameters import (
    SHParameters,
    supports_cloud_coverage,
    get_default_parameters,
)
from .evalscripts import EvalScripts
from .utils import (
    evaluate_sentinel_parameters,
    imagery_from_tmp_to_dir,
    filter_times,
    stats_to_df,
)



import torch 
import rasterio as rio
import numpy as np
from skimage import io


class Dataset(torch.utils.data.Dataset):
    def __init__(self, images, trans, bands, norm_value=4000):
        self.images = images
        self.trans = trans
        self.bands = bands
        self.norm_value = norm_value

    def __len__(self):
        return len(self.images)
    
    def __getitem__(self, idx):
        with rio.open(self.images[idx]) as image_src:
            image = np.clip(image_src.read(self.bands) / self.norm_value, 0, 1).astype(np.float32).transpose(1, 2, 0)
        if self.trans:
            # get two random augmented version of image 
            # random cropping, resizing, horizontal flipping, color jittering, converting to grayscale, Gaussian blurring, and solarization
            image1 = self.trans(image=image)['image']
            image2 = self.trans(image=image)['image']
        return torch.from_numpy(image1).float().permute(2, 0, 1), torch.from_numpy(image2).float().permute(2, 0, 1)
    
class EuroSATDataset(torch.utils.data.Dataset):
    def __init__(self, images, labels, bands, norm_value=4000):
        self.images = images
        self.labels = labels
        self.bands = [b - 1 for b in bands]
        self.norm_value = norm_value
    
    def __len__(self):
        return len(self.images)

    def __getitem__(self, ix):
        img = io.imread(self.images[ix])[..., self.bands]
        img = torch.from_numpy(
            img / self.norm_value).clip(0, 1).float().permute(2, 0, 1)
        label = torch.tensor(self.labels[ix], dtype=torch.long)
        return img, label



import pytest
import numpy as np
import rasterio
import os
import shutil
import pandas as pd

from sentinelhub import BBox

from eotdl.tools.geo_utils import *
from shapely.geometry import box, Point


@pytest.mark.parametrize("bbox, expected", [
    ((1, 2, 3, 4), True),
    ((-10, -20, 10, 20), True),
    ((0, 0, 100, 100), True),
    ((1, 2, 3), False),
    ((3, 2, 1, 4), False),
    ((1, 4, 3, 2), False),
    (("a", "b", "c", "d"), False),
    ((1, 2, 3, 4, 5), False),
])
def test_is_bounding_box(bbox, expected):
    assert is_bounding_box(bbox) == expected


@pytest.mark.parametrize("input_bbox, expected_output", [
    ([1, 2, 3, 4], box(1, 2, 3, 4)),
])
def test_bbox_to_polygon(input_bbox, expected_output):
    assert bbox_to_polygon(input_bbox) == expected_output


class ParametersMock:
    RESOLUTION = 10


def test_compute_image_size():
    bounding_box = [13.35, 52.47, 13.39, 52.5]
    parameters = ParametersMock()

    bbox, bbox_size = compute_image_size(bounding_box, parameters)

    assert isinstance(bbox, BBox)
    assert isinstance(bbox_size, tuple)
    assert len(bbox_size) == 2
    assert bbox_size == (279, 328)


@pytest.fixture
def mock_raster():
    os.makedirs("tmp", exist_ok=True)
    raster_path = "tmp/mock_raster.tif"

    data = np.random.rand(100, 100).astype('float32')

    with rasterio.open(
        raster_path,
        'w',
        driver='GTiff',
        height=data.shape[0],
        width=data.shape[1],
        count=1,
        dtype=data.dtype,
        crs='+proj=latlong',
        transform=rasterio.transform.from_origin(1, 1, 0.05, 0.05),
    ) as dst:
        dst.write(data, 1)

    yield raster_path
    shutil.rmtree("tmp")


def test_get_image_bbox(mock_raster):
    bbox = get_image_bbox(mock_raster)
    assert isinstance(bbox, list)
    assert len(bbox) == 4


def test_get_image_resolution(mock_raster):
    resolution = get_image_resolution(mock_raster)
    assert isinstance(resolution, tuple)
    assert resolution == (0.05, 0.05)


def test_bbox_to_coordinates():
    bbox = [1, 2, 3, 4]
    coordinates = bbox_to_coordinates(bbox)
    assert isinstance(coordinates, list)
    assert len(coordinates) == 5
    assert coordinates == [(1, 2), (1, 4), (3, 4), (3, 2), (1, 2)]


def test_bbox_to_polygon():
    bbox = [1, 2, 3, 4]
    polygon = bbox_to_polygon(bbox)
    assert isinstance(polygon, Polygon)
    assert polygon.bounds == (1, 2, 3, 4)


def test_bbox_from_centroid():
    bbox = bbox_from_centroid(0, 0, 10, 2, 2)
    assert isinstance(bbox, list)
    assert len(bbox) == 4
    bbox_round = [round(i, 4) for i in bbox]
    assert bbox_round == [-0.0001, -0.0001, 0.0001, 0.0001]


def test_generate_bounding_box():
    point = Point(0, 0)
    differences = [2, 2]

    bbox = generate_bounding_box(point, differences)
    expected_bbox = [-1, -1, 1, 1]

    assert bbox == expected_bbox


def test_convert_df_geom_to_shape():
    data = {'geometry': [Point(0, 0).__geo_interface__]}
    df = pd.DataFrame(data)

    row_with_geom = df.iloc[0]
    result_wkt_with_geom = convert_df_geom_to_shape(row_with_geom)

    assert result_wkt_with_geom == "POINT (0 0)"

    row_without_geom = pd.Series({"geometry": None})
    result_wkt_without_geom = convert_df_geom_to_shape(row_without_geom)

    assert result_wkt_without_geom == "POLYGON EMPTY"



import geopandas as gpd
from shapely.geometry import Point

def buffer_geometry(gdf: gpd.GeoDataFrame, buffer: int, resolution: int) -> gpd.GeoDataFrame:
    """
    Buffer the geometries in a GeoDataFrame and return the modified GeoDataFrame.
    
    Parameters:
        gdf: Input GeoDataFrame with geometries to buffer.
        buffer: Buffer distance in meters.
    
    Returns:
        A GeoDataFrame with buffered geometries.
    """
    # Ensure the GeoDataFrame has a valid CRS
    if gdf.crs is None:
        raise ValueError("Input GeoDataFrame must have a defined CRS.")

    # Estimate the UTM CRS based on the data's centroid
    utm = gdf.estimate_utm_crs()
    gdf_utm = gdf.to_crs(utm)
    
    # Round the centroids to the nearest 20m grid and apply buffering
    gdf_utm['geometry'] = gdf_utm.centroid.apply(
        lambda point: Point(round(point.x / resolution) * resolution, round(point.y / resolution) * resolution)
    ).buffer(distance=buffer, cap_style=3)
    
    return gdf_utm



# src/cbs/ssl_online_evaluator.py

import torch
from lightning.pytorch.callbacks import Callback
from tqdm.auto import tqdm
from .dm import EuroSATDataModule


class SSLOnlineEvaluator(Callback):
    """
    Every `eval_period` epochs, freeze the backbone, train a small linear head
    (for `head_epochs` epochs) on EuroSAT's train split, then evaluate on EuroSAT's
    val split and log accuracy. Additionally, run once before training epoch 1.

    Args:
        path (str): where EuroSAT lives on disk
        bands (tuple): which bands to load, e.g. (4,3,2,8)
        eval_period (int): run linear-eval every `eval_period` epochs
        head_epochs (int): how many epochs to train the linear head each time
        num_classes (int): number of classes (EuroSAT is 10)
        lr (float): learning rate for the linear head
    """

    def __init__(
        self,
        path: str,
        bands: tuple = (4, 3, 2, 8),
        label_ratio: float = 0.01,
        eval_period: int = 10,
        head_epochs: int = 10,
        num_classes: int = 10,
        lr: float = 1e-3,
    ):
        super().__init__()
        self.eval_period = eval_period
        self.head_epochs = head_epochs
        self.num_classes = num_classes
        self.lr = lr
        self.label_ratio = label_ratio

        # Our EuroSAT DataModule (for the "linear eval" part)
        self.dm = EuroSATDataModule(path=path, bands=bands, label_ratio=self.label_ratio)

        # placeholders; set in on_train_start
        self.head: torch.nn.Module = None
        self.optimizer: torch.optim.Optimizer = None
        self.criterion = torch.nn.CrossEntropyLoss()
        self.feat_dim: int = None

    def on_train_start(self, trainer, pl_module):
        """
        Called once, before the very first training epoch (epoch 0) begins.
        We:
          1) call dm.setup()
          2) do a quick forward pass to infer feature dims → build self.head + self.optimizer
          3) run the linear-evaluation right away (helper method)
        """
        # ─── 1) prepare EuroSAT splits ─────────────────────────────────
        self.dm.setup(stage="fit")

        # ─── 2) build the linear head ──────────────────────────────────
        # grab one batch just to infer "C" from the backbone's output
        train_loader = self.dm.train_dataloader()
        inputs, _ = next(iter(train_loader))
        inputs = inputs.to(pl_module.device)

        backbone = pl_module.backbone
        backbone.eval()
        with torch.no_grad():
            feats = backbone(inputs)
            if isinstance(feats, (list, tuple)):
                feats = feats[-1]
        self.feat_dim = feats.shape[1] # Store feat_dim

        # Remove head and optimizer initialization from here
        # self.head = torch.nn.Sequential(
        #     torch.nn.AdaptiveAvgPool2d((1, 1)),
        #     torch.nn.Flatten(),
        #     torch.nn.Linear(feat_dim, self.num_classes),
        # ).to(pl_module.device)

        # self.optimizer = torch.optim.Adam(self.head.parameters(), lr=self.lr)

        # return the backbone to train mode (so your main unsupervised loss still updates it)
        backbone.train()

        # ─── 3) run a first evaluation "before epoch 1" ────────────────
        # By calling our helper method here, we get a "first peek" at
        # `"eurosat_val_acc"` right away, before any training epoch runs.
        self._run_linear_eval(trainer, pl_module)

    def on_train_epoch_end(self, trainer, pl_module, outputs=None):
        """
        Called after each training epoch. If (epoch+1) % eval_period == 0,
        do another linear evaluation.
        """
        epoch = trainer.current_epoch
        # keep the same periodic check:
        if (epoch + 1) % self.eval_period != 0:
            return

        self._run_linear_eval(trainer, pl_module)

    def _run_linear_eval(self, trainer, pl_module):
        """
        Helper that:
          1) freezes the backbone
          2) trains self.head for head_epochs on EuroSAT train split (with a tqdm bar)
          3) evaluates self.head on EuroSAT val split (computes accuracy)
          4) logs "eurosat_val_acc" via pl_module.log(...)
          5) unfreezes the backbone
        """
        backbone = pl_module.backbone
        device = pl_module.device

        # ─── 1) freeze backbone ───────────────────────────────────────
        backbone.eval()
        for p in backbone.parameters():
            p.requires_grad = False

        if trainer.is_global_zero:
            # Re-initialize head and optimizer here to train from scratch
            self.head = torch.nn.Sequential(
                torch.nn.AdaptiveAvgPool2d((1, 1)),
                torch.nn.Flatten(),
                torch.nn.Linear(self.feat_dim, self.num_classes),
            ).to(device)
            self.optimizer = torch.optim.Adam(self.head.parameters(), lr=self.lr)

            # ─── 2) train the head for head_epochs, showing a tqdm bar ────
            self.head.train()
            # print(f"Training head for {self.head_epochs} epochs...", end="", flush=True)
            for head_epoch in range(self.head_epochs):
                # loop = tqdm(
                #     self.dm.train_dataloader(),
                #     desc=f"[Head {head_epoch+1}/{self.head_epochs}]",
                #     leave=False,
                # )
                # for inputs, targets in loop:
                for inputs, targets in self.dm.train_dataloader():
                    inputs = inputs.to(pl_module.device)
                    targets = targets.to(pl_module.device)

                    with torch.no_grad():
                        feats = backbone(inputs)
                        if isinstance(feats, (list, tuple)):
                            feats = feats[-1]
                    feats = feats.detach()

                    logits = self.head(feats)
                    loss = self.criterion(logits, targets)

                    self.optimizer.zero_grad()
                    loss.backward()
                    self.optimizer.step()

                    # loop.set_postfix(loss=loss.item())

            # ─── 3) evaluate the head on EuroSAT validation ───────────────
            self.head.eval()
            correct = 0
            total = 0
            for inputs, targets in self.dm.val_dataloader():
                inputs = inputs.to(pl_module.device)
                targets = targets.to(pl_module.device)

                with torch.no_grad():
                    feats = backbone(inputs)
                    if isinstance(feats, (list, tuple)):
                        feats = feats[-1]
                feats = feats.detach()
                logits = self.head(feats)
                preds = torch.argmax(logits, dim=1)

                correct += (preds == targets).sum().item()
                total += targets.size(0)

            val_acc = correct / total

            # ─── 4) log accuracy to Lightning's logger & progress bar ────
            #   - on_epoch=True ensures it shows up in the epoch-level metrics
            #   - prog_bar=True makes it appear in the tqdm bar
            pl_module.log(
                "eurosat_val_acc",
                val_acc,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                sync_dist=True,
            )

        if trainer.world_size > 1:
            torch.distributed.barrier()

        # ─── 5) unfreeze backbone ─────────────────────────────────────
        backbone.train()
        for p in backbone.parameters():
            p.requires_grad = True



import torch.nn as nn
import torch

from ..segmentation.model import SegmentationModel

class MultiTaskModel(SegmentationModel):
    def __init__(self, num_seg_classes=6, num_cls_classes=6, encoder='resnet18', in_channels=3, pretrained=True):
        super().__init__(num_seg_classes, encoder, in_channels, pretrained)
        self.classifier = nn.Sequential(
            torch.nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            torch.nn.Flatten(),
            torch.nn.Linear(
                self.encoder.feature_info.channels(-1), num_cls_classes)
        )

    def forward(self, x):
        f = self.encoder(x)
        x = f[-1]
        for i, d in enumerate(self.decoder):
            x = d(x, f[-i-2])
        return self.conv_last(x), self.classifier(f[-1])


from pathlib import Path

from ..repos import FEAPIRepo
from ..files.ingest import prep_ingest_folder, ingest

def retrieve_pipeline(metadata, user, private=False):
	repo = FEAPIRepo()
	data, error = repo.retrieve_pipeline(metadata.name)
	if data and data["uid"] != user["uid"]:
		raise Exception("Pipeline already exists.")
	if error and error == "Pipeline doesn't exist":
		# create pipeline
		data, error = repo.create_pipeline(metadata.dict(), user)
		# print(data, error)
		if error:
			raise Exception(error)
	return data

def ingest_openeo(
	path,
	verbose=False,
	logger=print,
	force_metadata_update=False,
	sync_metadata=False,
	private=False,
):
	path = Path(path)
	if not path.is_dir():
		raise Exception("Path must be a folder")
	# if "catalog.json" in [f.name for f in path.iterdir()]:
	# 	prep_ingest_stac(path, logger)
	# else:
	# 	prep_ingest_folder(path, verbose, logger, force_metadata_update, sync_metadata)
	prep_ingest_folder(path, verbose, logger, force_metadata_update, sync_metadata)
	return ingest(path, FEAPIRepo(), retrieve_pipeline, 'pipelines', private)


# def ingest_virtual_dataset( # could work for a list of paths with minimal changes...
# 	path,
# 	links,
# 	metadata = None, 
# 	logger=print,
# 	user=None,
# ):
# 	return ingest_virtual(path, links, DatasetsAPIRepo(), retrieve_dataset, 'datasets', metadata, logger)

# def ingest_dataset_catalog(path, logger=None):
# 	path = Path(path)
# 	return ingest_catalog(path, DatasetsAPIRepo(), retrieve_dataset, 'datasets')


from __future__ import (
    annotations,  # Required because of type annotations for return results
)

import boto3
from boto3.s3.transfer import TransferConfig
from dataclasses import dataclass
import os
import hashlib
import datetime
import geopandas as gpd
from openeo.rest.connection import Connection
from tempfile import NamedTemporaryFile

def upload_geoparquet_artifactory(gdf: gpd.GeoDataFrame, row_id: int) -> str:
    # Save the dataframe as geoparquet to upload it to artifactory
    temporary_file = NamedTemporaryFile()
    gdf.to_parquet(temporary_file.name)



@dataclass(frozen=True)
class S3URI:
    bucket: str
    key: str

    @classmethod
    def from_str(cls, uri: str) -> S3URI:
        s3_prefix = "s3://"
        if uri.startswith(s3_prefix):
            without_prefix = uri[len(s3_prefix) :]
            without_prefix_parts = without_prefix.split("/")
            bucket = without_prefix_parts[0]
            if len(without_prefix_parts) == 1:
                return S3URI(bucket, "")
            else:
                return S3URI(bucket, "/".join(without_prefix_parts[1:]))
        else:
            raise ValueError(
                "Input {uri} is not a valid S3 URI should be of form s3://<bucket>/<key>"
            )


@dataclass(frozen=True)
class AWSSTSCredentials:
    AWS_ACCESS_KEY_ID: str
    AWS_SECRET_ACCESS_KEY: str
    AWS_SESSION_TOKEN: str
    subject_from_web_identity_token: str
    STS_ENDPOINT = "https://sts.prod.warsaw.openeo.dataspace.copernicus.eu"

    @classmethod
    def _from_assume_role_response(cls, resp: dict) -> AWSSTSCredentials:
        d = resp["Credentials"]

        return AWSSTSCredentials(
            AWS_ACCESS_KEY_ID=d["AccessKeyId"],
            AWS_SECRET_ACCESS_KEY=d["SecretAccessKey"],
            AWS_SESSION_TOKEN=d["SessionToken"],
            subject_from_web_identity_token=resp["SubjectFromWebIdentityToken"],
        )

    def set_as_environment_variables(self) -> None:
        """If temporary credentials are to be used elsewhere in the notebook"""
        os.environ["AWS_ACCESS_KEY_ID"] = self.AWS_ACCESS_KEY_ID
        os.environ["AWS_SECRET_ACCESS_KEY"] = self.AWS_SECRET_ACCESS_KEY
        os.environ["AWS_SESSION_TOKEN"] = self.AWS_SESSION_TOKEN

    def as_kwargs(self) -> dict:
        return {
            "aws_access_key_id": self.AWS_ACCESS_KEY_ID,
            "aws_secret_access_key": self.AWS_SECRET_ACCESS_KEY,
            "aws_session_token": self.AWS_SESSION_TOKEN,
        }

    @classmethod
    def from_openeo_connection(cls, conn: Connection) -> AWSSTSCredentials:
        """
        Takes an OpenEO connection object and returns temporary credentials to interact with S3
        """
        auth_token = conn.auth.bearer.split("/")
        os.environ["AWS_ENDPOINT_URL_STS"] = cls.STS_ENDPOINT
        sts = boto3.client("sts")
        return AWSSTSCredentials._from_assume_role_response(
            sts.assume_role_with_web_identity(
                RoleArn="arn:aws:iam::000000000000:role/S3Access",
                RoleSessionName=auth_token[1],
                WebIdentityToken=auth_token[2],
                DurationSeconds=43200,
            )
        )

    def get_user_hash(self) -> str:
        hash_object = hashlib.sha1(self.subject_from_web_identity_token.encode())
        return hash_object.hexdigest()


class OpenEOArtifactHelper:
    BUCKET_NAME = "OpenEO-artifacts"
    S3_ENDPOINT = "https://s3.prod.warsaw.openeo.dataspace.copernicus.eu"
    # From what size will we switch to multi-part-upload
    MULTIPART_THRESHOLD_IN_MB = 50

    def __init__(self, creds: AWSSTSCredentials):
        self._creds = creds
        self.session = boto3.Session(**creds.as_kwargs())

    @classmethod
    def from_openeo_connection(cls, conn: Connection) -> OpenEOArtifactHelper:
        creds = AWSSTSCredentials.from_openeo_connection(conn)
        return OpenEOArtifactHelper(creds)

    def get_s3_client(self):
        return self.session.client("s3", endpoint_url=self.S3_ENDPOINT)

    def set_env(self):
        os.environ["AWS_ENDPOINT_URL_S3"] = self.S3_ENDPOINT

    def user_prefix(self) -> str:
        """Each user has its own prefix retrieve it"""
        return self._creds.get_user_hash()

    def get_upload_prefix(self):
        return (
            f"{self.user_prefix()}/{datetime.datetime.utcnow().strftime('%Y/%m/%d')}/"
        )

    def get_upload_key(self, object_name: str) -> str:
        return f"{self.get_upload_prefix()}{object_name}"

    def upload_bytes(self, object_name: str, blob: bytes) -> str:
        """Upload a bunch of bytes into an object and return an S3 URI to it"""
        bucket = self.BUCKET_NAME
        key = self.get_upload_key(object_name)
        self.get_s3_client().put_object(Body=blob, Bucket=bucket, Key=key)
        return f"s3://{bucket}/{key}"

    def upload_string(self, object_name: str, s: str, encoding: str = "utf-8") -> str:
        """Upload a string into an object"""
        return self.upload_bytes(object_name, s.encode(encoding))

    def upload_file(self, object_name: str, src_file_path: str) -> str:
        MB = 1024**2
        config = TransferConfig(multipart_threshold=self.MULTIPART_THRESHOLD_IN_MB * MB)

        bucket = self.BUCKET_NAME
        key = self.get_upload_key(object_name)

        self.get_s3_client().upload_file(src_file_path, bucket, key, Config=config)
        return f"s3://{bucket}/{key}"

    def get_presigned_url(
        self, s3_uri: str, expires_in_seconds: int = 3600 * 24 * 6
    ) -> str:
        typed_s3_uri = S3URI.from_str(s3_uri)
        return self.get_s3_client().generate_presigned_url(
            "get_object",
            Params={"Bucket": typed_s3_uri.bucket, "Key": typed_s3_uri.key},
            ExpiresIn=expires_in_seconds,
        )
    


def upload_geoparquet_file(gdf: gpd.GeoDataFrame, conn: Connection) -> str:
    geo_parquet_path = "bounding_box_geometry.parquet"
    gdf.to_parquet(geo_parquet_path)

    # Create an instance of the custom S3 uploader
    uploader = OpenEOArtifactHelper.from_openeo_connection(conn)

    # Upload the GeoParquet file to S3
    s3_uri = uploader.upload_file(geo_parquet_path, geo_parquet_path)

    # Get the presigned URL for accessing the uploaded file
    presigned_url = uploader.get_presigned_url(s3_uri)

    # Optionally clean up the local file
    os.remove(geo_parquet_path)

    return presigned_url


import typer
from ..auth import is_logged, auth, logout_user
from ..auth.errors import LoginError

app = typer.Typer(help="Login to EOTDL.")


@app.command()
def login():
    """
    Login to the EOTDL.

    This command will return a URL that you can visit to authenticate.
    After authentication, your credentials will be stored locally.
    This enables future commands to be executed without having to authenticate again
    (at least while the credentials are valid).
    \n\n
    The default browser will be opened automatically. If not, you can copy the URL and paste it in your browser.
    \n\n
    By default, the credentials will be stored in the following file: ~/.eotdl/credentials.json
    """
    try:
        user = auth()
        typer.echo(f"You are logged in as {user['email']}")
    except LoginError as e:
        typer.echo(e.message)
        raise typer.Exit(code=1)


@app.command()
def logout(
    yes: bool = typer.Option(False, "--yes", "-y", help="Skip confirmation")
):
    """
    Logout from the EOTDL.

    You will receive a logout url that you can visit in case you want to authenticate with a different account.
    """
    user = is_logged()
    if user:
        typer.echo(f"You are logged in as {user['email']}")
        if not yes:
            typer.confirm("Are you sure you want to logout?", abort=True)
        logout_url = logout_user()
        typer.echo("You are logged out.")
        typer.echo(
            f"If you want to login with a different account, visit {logout_url} and login again."
        )
    else:
        typer.echo("You are not logged in.")


if __name__ == "__main__":
    app()



import typer
from pathlib import Path

from ..datasets import (
    retrieve_datasets,
    ingest_dataset,
    stage_dataset,
    deactivate_dataset,
    retrieve_private_datasets
)

app = typer.Typer(help="Explore, ingest and download training datasets.")

@app.command()
def ingest(
    path: Path = typer.Option(
        ..., "--path", "-p", help="Path to the dataset to ingest"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Verbose output. This will print the progress of the ingestion",
    ),
    foce_metadata_update: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force metadata update even if it already exists. Will overwrite the current metadata in EOTDL",
    ),
    sync_metadata: bool = typer.Option(
        False,
        "--sync",
        "-s",
        help="Sync local metadata with the EOTDL. Will overwrite the local metadata",
    ),
    private: bool = typer.Option(
        False,
        '--private',
        '-pr',
        help="Make dataset private"
    ),
    ignore_stac: bool = typer.Option(
        False,
        "--ignore-stac",
        "-is",
        help="Ignore STAC catalog.json (if found) for geneating metadata."
    )
):
    """
    Ingest a dataset to the EOTDL.asdf

    This command ingests the dataset to the EOTDL. The dataset must be a folder with the dataset files,
    and at least a README.md file (and a catalog.json file for Q1+). If these files are missing, the ingestion
    will not work. All the files in the folder will be uploaded to the EOTDL.
    \n\n
    The following constraints apply to the dataset name:\n
    - It must be unique\n
    - It must be between 3 and 45 characters long\n
    - It can only contain alphanumeric characters and dashes.\n
    \n
    The README.md file must contain the following fields in the metadata header:\n
    - name: the name of the dataset\n
    - authors: the author or authors of the dataset\n
    - license: the license of the dataset\n
    - source: the source of the dataset\n
    - thumbnail: an image to use as the thumbnail of the dataset in the website\n
    The rest of the content in the README.md file will be used as the description of the dataset in the website.
    If using --verbose, it will print the progress of the ingestion.
    \n\n
    Examples\n
    --------\n
    $ eotdl dataset ingest --path /path/to/folder-with-dataset --verbose True
    """
    try:
        ingest_dataset(path, verbose, typer.echo, foce_metadata_update, sync_metadata, private, ignore_stac)
    except Exception as e:
        typer.echo(e)
        
@app.command()
def list(
    name: str = typer.Option(
        None, "--name", "-n", help="Filter the returned datasets by name"
    ),
    limit: int = typer.Option(
        None, "--limit", "-l", help="Limit the number of returned results"
    ),
    private: bool = typer.Option(
        False, "--private", "-p", help="Show private datasets"
    ),
):
    """
    Retrieve a list with all the datasets in the EOTDL.

    If using --name, it will filter the results by name. If no name is provided, it will return all the datasets.\n
    If using --limit, it will limit the number of results. If no limit is provided, it will return all the datasets.
    \n\n
    Examples\n
    --------\n
    $ eotdl datasets list\n
    $ eotdl datasets list --name YourModel --limit 5
    """
    try:
        if private:
            datasets = retrieve_private_datasets()
        else:
            datasets = retrieve_datasets(name, limit)
        typer.echo(datasets)
    except Exception as e:
        typer.echo(e)

@app.command()
def get(
    dataset: str = typer.Argument(None, help="Name of the dataset to download"),
    path: str = typer.Option(
        None, "--path", "-p", help="Download the dataset to a specific output path"
    ),
    version: int = typer.Option(None, "--version", "-v", help="Dataset version"),
    assets: bool = typer.Option(
        False, "--assets", "-a", help="Download STAC assets from the dataset"
    ),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force download even if file exists"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the download",
    ),
):
    """
    Download a dataset from the EOTDL.
    \n\n
    If using --path, it will download the dataset to the specified path. If no path is provided, it will download to ~/.eotdl/datasets.\n
    If using --version, it will download the specified version. If no version is provided, it will download the latest version.\n
    If using --assets when the dataset is STAC, it will also download the STAC assets of the dataset. If not provided, it will only download the STAC metadata.\n
    If using --force, it will download the dataset even if the file already exists.\n
    If using --verbose, it will print the progress of the download.\n
    \n\n
    Examples\n
    --------\n
    $ eotdl dataset get YourDataset\n
    $ eotdl dataset get YourDataset --path /path/to/download --version 1 --assets True --force True --verbose True
    """
    try:
        dst_path = stage_dataset(
            dataset,
            version,
            path,
            typer.echo,
            assets,
            force,
            verbose,
        )
        typer.echo(f"Data available at {dst_path}")
    except Exception as e:
        typer.echo(e)


@app.command()
def delete(
    dataset: str = typer.Argument(None, help="Name of the dataset to deactivate")
):
    """
    Delete a dataset from the EOTDL.
    """
    try:
        deactivate_dataset(dataset)
        typer.echo(f"Dataset {dataset} deleted")
    except Exception as e:
        typer.echo(e)

if __name__ == "__main__":
    app()



from glob import glob
import json
import os
from torch.utils.data import DataLoader
import lightning as L

from .ds import SegmentationDataset, BinarySegmentationDataset
from ..classification.dm import SCANEOClassificationDataModule
    
class SCANEOSegmentationDataModule(SCANEOClassificationDataModule):
    def __init__(self, 
            path, 
            batch_size=16, 
            num_workers=4, 
            pin_memory=True, 
            val_split=0.2, 
            train_trans=None, 
            val_trans=None
        ):
        super().__init__(
            path=path,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            val_split=val_split,
            train_trans=train_trans,
            val_trans=val_trans
        )
        with open(f'{path}/spai.json', 'r') as f:
            self.spai_labels = json.load(f)
        self.classes = [label['name'] for label in self.spai_labels['labels']]
        self.num_classes = len(self.classes) + 1
        self.colors = {label['name']: label['color'] for label in self.spai_labels['labels']}

    def setup(self, stage=None):
        masks = sorted(glob(f'{self.path}/*_mask.tif'))
        images = [i.replace('_mask.tif', '.tif') for i in masks]
        val_size = int(len(images) * self.val_split)
        train_image = images[:-val_size]
        val_image = images[-val_size:]
        train_mask = masks[:-val_size]
        val_mask = masks[-val_size:]
        print("Training on", len(train_image), "images")
        print("Validating on", len(val_image), "images")
        self.train_ds = SegmentationDataset(train_image, train_mask, trans=self.train_trans, num_classes=self.num_classes)
        self.val_ds = SegmentationDataset(val_image, val_mask, trans=self.val_trans, num_classes=self.num_classes)

class SegmentationDataModule(L.LightningDataModule):
    def __init__(self, root, train_source_path, val_source_path, train_labels_path, val_labels_path, test_source_path=None, test_labels_path=None, batch_size=5, num_workers=4, pin_memory=True, train_trans=None, val_trans=None, test_trans=None, bands=(4,3,2), norm_value=4000, source_ext=".tiff", label_ext=".tif"):
        super().__init__()
        self.root = root
        self.train_source_path = train_source_path
        self.val_source_path = val_source_path
        self.test_source_path = test_source_path
        self.train_labels_path = train_labels_path
        self.val_labels_path = val_labels_path
        self.test_labels_path = test_labels_path  
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.train_trans = train_trans
        self.val_trans = val_trans
        self.test_trans = test_trans
        self.bands = bands
        self.norm_value = norm_value
        self.source_ext = source_ext
        self.label_ext = label_ext

    def setup(self, stage=None):
        train_images = sorted(os.listdir(os.path.join(self.root, self.train_source_path)))
        val_images = sorted(os.listdir(os.path.join(self.root, self.val_source_path)))
        if self.test_source_path is not None:
            test_images = sorted(os.listdir(os.path.join(self.root, self.test_source_path)))
        train_images_paths = [os.path.join(self.root, self.train_source_path, image) for image in train_images]
        val_images_paths = [os.path.join(self.root, self.val_source_path, image) for image in val_images]
        if self.test_source_path is not None:
            test_images_paths = [os.path.join(self.root, self.test_source_path, image) for image in test_images]
        train_labels_paths = [os.path.join(self.root, self.train_labels_path, self.change_ext(image, self.source_ext, self.label_ext)) for image in train_images]
        val_labels_paths = [os.path.join(self.root, self.val_labels_path, self.change_ext(image, self.source_ext, self.label_ext)) for image in val_images]
        test_labels_paths = [os.path.join(self.root, self.test_labels_path, self.change_ext(image, self.source_ext, self.label_ext)) for image in test_images]
        self.train_ds = BinarySegmentationDataset(train_images_paths, train_labels_paths, self.train_trans, self.bands, self.norm_value)
        self.val_ds = BinarySegmentationDataset(val_images_paths, val_labels_paths, self.val_trans, self.bands, self.norm_value)
        self.test_ds = BinarySegmentationDataset(test_images_paths, test_labels_paths, self.test_trans, self.bands, self.norm_value)

    def change_ext(self, p, ext1, ext2):
        if ext1 != ext2:
            return p.replace(ext1, ext2) 
        return p

    def train_dataloader(self, batch_size=None, shuffle=True):
        return DataLoader(self.train_ds, batch_size=batch_size or self.batch_size, shuffle=shuffle, num_workers=self.num_workers, pin_memory=self.pin_memory)

    def val_dataloader(self, batch_size=None, shuffle=False):
        return DataLoader(self.val_ds, batch_size=batch_size or self.batch_size, shuffle=shuffle, num_workers=self.num_workers, pin_memory=self.pin_memory)
    
    def test_dataloader(self, batch_size=None, shuffle=False):
        return DataLoader(self.test_ds, batch_size=batch_size or self.batch_size, shuffle=shuffle, num_workers=self.num_workers, pin_memory=self.pin_memory)


import lightning as L
import numpy as np
import albumentations as A
from torch.utils.data import DataLoader

from .ds import Dataset

class ESRTDatamodule(L.LightningDataModule):
	def __init__(
			self, 
			hr_paths, 
			lr_paths, 
			upscale = 2,
			batch_size = 16, 
			val_size = 0.2,
			seed = 2025,
			num_workers = 20,
			pin_memory = True,
			train_samples = None,
			resize = True,
			hr_bands=(0, 1, 2, 3), 	
			lr_bands=(3, 2, 1, 7),
		):
		super().__init__()
		self.hr_paths = hr_paths
		self.lr_paths = lr_paths
		self.batch_size = batch_size
		self.num_workers = num_workers
		self.val_size = val_size
		self.seed = seed
		self.num_workers = num_workers
		self.pin_memory = pin_memory
		self.upscale = upscale
		self.train_samples = train_samples
		self.resize = resize
		self.hr_bands = hr_bands
		self.lr_bands = lr_bands

	def setup(self, stage = None):
		np.random.seed(self.seed)
		assert len(self.hr_paths) == len(self.lr_paths), "Number of hr and lr images must be the same"
		if self.train_samples is not None:
			self.hr_paths = self.hr_paths[:self.train_samples]
			self.lr_paths = self.lr_paths[:self.train_samples]
		ixs = np.arange(len(self.hr_paths))
		np.random.shuffle(ixs)
		val_size = int(len(ixs) * self.val_size)
		train_ixs = ixs[:-val_size]
		val_ixs = ixs[-val_size:]
		hr_train_paths, hr_val_paths = np.array(self.hr_paths)[train_ixs], np.array(self.hr_paths)[val_ixs]
		lr_train_paths, lr_val_paths = np.array(self.lr_paths)[train_ixs], np.array(self.lr_paths)[val_ixs]
		print(f"Using {len(hr_train_paths)} training images and {len(hr_val_paths)} validation images")
		assert len(hr_train_paths) == len(lr_train_paths), "Number of hr and lr images must be the same"
		assert len(hr_val_paths) == len(lr_val_paths), "Number of hr and lr images must be the same"
		trans = A.Compose([
			A.HorizontalFlip(),
			A.VerticalFlip(),
			A.Transpose(),
			A.RandomRotate90(),
		], additional_targets={"image2": "image"}, is_check_shapes=False)
		self.train_ds = Dataset(hr_train_paths, lr_train_paths, self.hr_bands, self.lr_bands,  self.upscale, trans, resize=self.resize)
		self.val_ds = Dataset(hr_val_paths, lr_val_paths, self.hr_bands, self.lr_bands, self.upscale, resize=self.resize)
		
	def get_dataloader(self, dataset, batch_size = None, shuffle = False):
		return DataLoader(
			dataset,
			batch_size=batch_size if batch_size is not None else self.batch_size,
			num_workers=self.num_workers,
			pin_memory=self.pin_memory,
			shuffle=shuffle if shuffle is not None else True
		)
	
	def train_dataloader(self, batch_size = None, shuffle = True):
		return self.get_dataloader(self.train_ds, batch_size, shuffle)
	
	def val_dataloader(self, batch_size = None, shuffle = False):
		return self.get_dataloader(self.val_ds, batch_size, shuffle)



import lightning as L
from glob import glob
from torch.utils.data import DataLoader
import albumentations as A
import os
from sklearn.model_selection import train_test_split
import numpy as np
import pandas as pd

from .ds import Dataset, EuroSATDataset

class ToGray(A.ImageOnlyTransform):
    def __init__(self, p=0.5):
        super().__init__(p=p)
        
    def apply(self, img, **params):
        gray = np.mean(img[:,:,:3], axis=2, keepdims=True)
        return np.repeat(gray, img.shape[2], axis=2).astype(img.dtype)

class UnsupervisedDataModule(L.LightningDataModule):
    def __init__(self, path, bands=(1,2,3,4), batch_size=256, num_workers=10, pin_memory=True, trans=None, norm_value=4000):
        super().__init__()
        self.path = path
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.trans = trans
        self.norm_value = norm_value
        self.bands = bands
        self.pin_memory = pin_memory

    def setup(self, stage=None):
        print("Unsupervised")
        images = glob(f'{self.path}/*')
        print(f"Number of images: {len(images)}")
        self.df = pd.DataFrame(images, columns=['image'])
        self.ds = Dataset(
            images,
            A.Compose([
                A.RandomResizedCrop(size=(224, 224), scale=(0.5, 1.0)),
                A.HorizontalFlip(p=0.5),
                A.VerticalFlip(p=0.5),
                A.Transpose(p=0.5),
                A.Rotate(p=0.5),
                A.ColorJitter(p=0.5), # expects RGB images # VERY IMPORTANT, WITHOUT THIS WILL NOT WORK (is there a version for RGBNir?)
                A.ToGray(p=0.3), # expects RGB images
                A.GaussianBlur(p=0.3),
            ]),
            self.bands,
            self.norm_value
        )

    def train_dataloader(self, batch_size=None, shuffle=True):
        return DataLoader(
            self.ds,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

class EuroSATDataModule(L.LightningDataModule):
    def __init__(self, path, bands=(4,3,2,8), norm_value=4000, batch_size=32, num_workers=10, label_ratio=1., pin_memory=True):
        super().__init__()
        self.path = path
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.label_ratio = label_ratio
        self.pin_memory = pin_memory
        self.bands = bands
        self.norm_value = norm_value

    def setup(self, stage=None):
        # get labels from folder structure if not provided
        self.labels = sorted(os.listdir(self.path))
        print("EuroSAT")
        print("Generating images and labels ...")
        images, encoded = [], []
        for ix, label in enumerate(self.labels):
            _images = os.listdir(f'{self.path}/{label}')
            images += [f'{self.path}/{label}/{img}' for img in _images]
            encoded += [ix]*len(_images)
        print(f'Number of images: {len(images)}')
        # train / val split
        print("Generating train / val splits ...")
        train_images, val_images, train_labels, val_labels = train_test_split(
            images,
            encoded,
            stratify=encoded,
            test_size=0.2,
            random_state=42
        )
        # filter by label ratio
        if self.label_ratio < 1. and self.label_ratio > 0.:
            train_images_ratio, train_labels_ratio = [], []
            unique_labels = np.unique(train_labels)
            for label in unique_labels:
                filter = np.array(train_labels) == label
                ixs = filter.nonzero()[0]
                num_samples = filter.sum()
                ratio_ixs = np.random.choice(
                    ixs, int(self.label_ratio*num_samples), replace=False)
                train_images_ratio += (np.array(train_images)
                                       [ratio_ixs]).tolist()
                train_labels_ratio += (np.array(train_labels)
                                       [ratio_ixs]).tolist()
            train_images = train_images_ratio
            train_labels = train_labels_ratio
        self.train_ds = EuroSATDataset(train_images, train_labels, self.bands, self.norm_value)
        self.val_ds = EuroSATDataset(val_images, val_labels, self.bands, self.norm_value)
        print("Training samples: ", len(self.train_ds))
        print("Validation samples: ", len(self.val_ds))

    def train_dataloader(self, batch_size=None, shuffle=True):
        return DataLoader(
            self.train_ds,
            batch_size=batch_size or self.batch_size,
            num_workers=self.num_workers,
            shuffle=shuffle,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self, batch_size=None, shuffle=False):
        return DataLoader(
            self.val_ds,
            batch_size=batch_size or self.batch_size,
            num_workers=self.num_workers,
            shuffle=shuffle,
            pin_memory=self.pin_memory
        )



from .dataframe_utils import *
import openeo
from openeo.extra.job_management import MultiBackendJobManager, CsvJobDatabase

# pass arguments in the row
def start_job(row: pd.Series, connection: openeo.Connection, **kwargs) -> openeo.BatchJob:
        temporal_extent = row["temporal_extent"]
        # set up load url in order to allow non-latlon feature collections for spatial filtering
        geometry = row["geometry"]
        #run the s1 and s2 udp
        s1 = connection.datacube_from_process(
                "s1_weekly_statistics", # depends on the json, so must be also a parameter
                namespace=row["s1_weekly_statistics_url"],
                temporal_extent=temporal_extent,
        )
        s2 = connection.datacube_from_process(
                "s2_weekly_statistics",
                namespace=row["s2_weekly_statistics_url"],
                temporal_extent=temporal_extent,
        )
        #merge both cubes and filter across the feature collection
        merged = s2.merge_cubes(s1)
        result = merged.aggregate_spatial(geometries = geometry, reducer = "mean")
        #dedicated job settings to save the individual features within a collection seperately
        job = result.create_job(
                out_format="parquet",
        )
        return job

def eurocrops_point_extraction(
        gdf,
        start_date,
        nb_months,
        extra_cols=[],
        job_tracker = 'jobs.csv',
        parallel_jobs=2,
):
        """
        # Transform GeoDataFrame for MultiBackendJobManager

        This function processes an input GeoDataFrame and prepares it for use with openEO's **MultiBackendJobManager**. The job manager enables launching and tracking multiple openEO jobs simultaneously, which is essential for large-scale data extractions. 

        ### Note

        It is important to note, that for this simple example we have opted to not group the various geometries into feature collections. This utility is only illustrated in the more advanced example. The impact for this choice is that for each polygon, a singly openEO job will need to be launched, leading to a more time and cost extensive extraction workflow.


        ### Parameters

        #### Temporal Parameters:
        - **Start Date:** Start of the temporal extent (e.g., `"2020-01-01"`).  
        - **Number of Months:** Duration of the temporal extent in months.
        """
        job_df = process_geodataframe(gdf, start_date, nb_months, extra_cols)
        """
        # Start Job with Standardized UDPs and Feature Collection Filtering

        This function initializes an openEO batch job using standardized **User-Defined Processes (UDPs)** for Sentinel-1 and Sentinel-2 data processing. It employs a spatial aggregation in order to get a time series per polygon.

        ### Key Features

        1. **Use of Standardized UDPs**  
        - **S1 Weekly Statistics:** Computes weekly statistics from Sentinel-1 data.  
        - **S2 Weekly Statistics:** Computes weekly statistics from Sentinel-2 data.  
        - UDPs are defined in external JSON files.

        2. **Spatial aggregation across polygons**  
        - an average is calculated for each individual polygon

        3. **Cube Merging**  
        - Merges Sentinel-1 and Sentinel-2 datacubes for combined analysis.

        4. **Job Configuration**  
        - Outputs results in **parquet** format with filenames derived
        """
        # Authenticate and add the backend
        connection = openeo.connect(url="openeo.dataspace.copernicus.eu").authenticate_oidc()
        # initialize the job manager
        manager = MultiBackendJobManager()
        manager.add_backend("cdse", connection=connection, parallel_jobs=parallel_jobs)
        job_db = CsvJobDatabase(path=job_tracker)
        if not job_db.exists():
                df = manager._normalize_df(job_df)
                job_db.persist(df)
        manager.run_jobs(start_job=start_job, job_db=job_db)


import rasterio as rio
import numpy as np
from torch.utils.data import Dataset
import torch

class ClassificationDataset(Dataset):
    def __init__(self, images, labels, bands=[1,2,3], norm_value = 4000, trans=None):
        self.images = images
        self.labels = labels
        self.trans = trans
        self.bands = bands
        self.norm_value = norm_value

    def __len__(self):
        return len(self.images)
    
    def __getitem__(self, idx):
        image = rio.open(self.images[idx]).read(self.bands)
        image = np.clip(image / self.norm_value, 0, 1).transpose(1,2,0)
        if self.trans is not None:
            image = self.trans(image=image)['image']
        labels = self.labels[idx]
        return torch.from_numpy(image.transpose(2,0,1)).float(), torch.from_numpy(labels).float()



"""
Utils for Sentinel Hub access
"""

import json

from os import makedirs
from datetime import datetime, timedelta
from typing import Union, Optional, Iterable, List
from glob import glob
from shutil import copyfile, rmtree
import pandas as pd

from .parameters import SUPPORTED_COLLECTION_IDS, SHParameters, OUTPUT_FORMAT
from ...tools.geo_utils import is_bounding_box
from ...tools.time_utils import is_time_interval, get_day_between
from sentinelhub import parse_time


def evaluate_sentinel_parameters(
    time_interval: Union[str, datetime],
    bounding_box: list,
    collection_id: Optional[str] = None,
    output: Optional[str] = None,
    output_needed: Optional[bool] = True,
    parameters: Optional[SHParameters] = None,
) -> None:
    """
    Evaluate parameters for Sentinel Hub access
    """
    if output_needed:
        if not output:
            raise ValueError("Output path must be specified.")
        if parameters and not parameters.OUTPUT_FORMAT:
            raise ValueError("Output format must be specified.")
    if collection_id:
        if collection_id not in SUPPORTED_COLLECTION_IDS:
            raise ValueError(
                f"Collection id {collection_id} is not supported. Supported collections ids are: {SUPPORTED_COLLECTION_IDS}"
            )
    else:
        if not (
            parameters
            and hasattr(parameters, "DATA_COLLECTION")
            and hasattr(parameters.DATA_COLLECTION, "api_id")
        ):
            raise ValueError(f"Data collection is not defined properly.")
    if not time_interval:
        raise ValueError("Time interval must be specified.")
    else:
        if len(time_interval) == 2 and not is_time_interval(time_interval):
            raise ValueError(
                "Time interval must be a list or tuple with two elements in format YYYY-MM-DD."
            )
    if not bounding_box:
        raise ValueError("Bounding box must be specified.")
    else:
        if not is_bounding_box(bounding_box):
            raise ValueError(
                "Bounding box must be a list or tuple with four elements in format (lon_min, lat_min, lon_max, lat_max)."
            )
    if parameters and parameters.MAX_CLOUD_COVERAGE:
        if not isinstance(parameters.MAX_CLOUD_COVERAGE, (int, float)) or (
            parameters.MAX_CLOUD_COVERAGE < 0 or parameters.MAX_CLOUD_COVERAGE > 100
        ):
            raise ValueError("Max cloud coverage must be a number between 0 and 100.")


def imagery_from_tmp_to_dir(
    output_dir: str,
    bounding_box: List[Union[int, float]],
    tmp_dir: Optional[str],
    name: Optional[str] = None,
    bulk: Optional[bool] = False,
    output_format: Optional[str] = OUTPUT_FORMAT.TIFF,
) -> None:
    """
    Copy imagery from tmp to output dir
    """
    format = output_format
    downloaded_files = glob(f"{tmp_dir}/**/response." + format)

    if len(downloaded_files) == 0:
        return
    makedirs(output_dir, exist_ok=True)
    for downloaded_file in downloaded_files:
        request_json = downloaded_file.replace("response." + format, "request.json")
        metadata = generate_raster_metadata(request_json, bounding_box)

        if name and not bulk:
            output_filename = name
        elif name and bulk:
            output_filename = f"{name}_{metadata['acquisition-date']}"
        else:
            if metadata["acquisition-date"]:
                output_filename = f"{metadata['type']}_{metadata['acquisition-date']}"
            else:
                output_filename = metadata["type"]
        copyfile(downloaded_file, f"{output_dir}/{output_filename}." + format)
        with open(f"{output_dir}/{output_filename}.json", "w", encoding="utf-8") as f:
            json.dump(metadata, f)
    rmtree(tmp_dir)


def generate_raster_metadata(request_json: str, bounding_box) -> None:
    """
    Generate metadata for raster
    """
    with open(request_json, "r", encoding="utf-8") as f:
        json_content = json.load(f)

    payload_data = json_content["request"]["payload"]["input"]["data"][0]
    sensor_type = payload_data["type"]
    # Get day between from and to
    if "timeRange" in payload_data["dataFilter"] and sensor_type != "dem":
        time_range = payload_data["dataFilter"]["timeRange"]
        acquisition_date = get_day_between(time_range["from"], time_range["to"])
    else:  # DEM data does not have a timeRange
        acquisition_date = None

    metadata = {
        "acquisition-date": acquisition_date,
        "bounding-box": bounding_box,
        "type": sensor_type,
    }

    return metadata


def filter_times(
    timestamps: Iterable[datetime], time_difference: timedelta
) -> list[datetime]:
    """
    Filters out timestamps within time_difference, preserving only the oldest timestamp.
    """
    timestamps = sorted(set(timestamps))

    filtered_timestamps: list[datetime] = []
    for current_timestamp in timestamps:
        if (
            not filtered_timestamps
            or current_timestamp - filtered_timestamps[-1] > time_difference
        ):
            filtered_timestamps.append(current_timestamp)

    return filtered_timestamps


def stats_to_df(stats_data):
    """Transform Statistical API response into a pandas.DataFrame"""
    df_data = []

    for single_data in stats_data["data"]:
        df_entry = {}
        is_valid_entry = True

        df_entry["interval_from"] = parse_time(single_data["interval"]["from"]).date()
        df_entry["interval_to"] = parse_time(single_data["interval"]["to"]).date()

        for output_name, output_data in single_data["outputs"].items():
            for band_name, band_values in output_data["bands"].items():
                band_stats = band_values["stats"]
                if band_stats["sampleCount"] == band_stats["noDataCount"]:
                    is_valid_entry = False
                    break

                for stat_name, value in band_stats.items():
                    col_name = f"{output_name}_{band_name}_{stat_name}"
                    if stat_name == "percentiles":
                        for perc, perc_val in value.items():
                            perc_col_name = f"{col_name}_{perc}"
                            df_entry[perc_col_name] = perc_val
                    else:
                        df_entry[col_name] = value

        if is_valid_entry:
            df_data.append(df_entry)

    return pd.DataFrame(df_data)



import lightning as L
from glob import glob
import geopandas as gpd
from torch.utils.data import DataLoader
import numpy as np
from sklearn.model_selection import train_test_split

from .ds import ClassificationDataset

def get_classification_labels(images, labels, classes):
    classification_labels = []
    remove = []
    for ix, label in enumerate(labels):
        _label = gpd.read_file(label)
        classification = _label[_label['task'] == 'classification']
        if len(classification) > 0:
            cls = classification.label.values
            cls = [classes.index(i) for i in cls if i in classes]
            _classes = np.zeros(len(classes), dtype=int)
            _classes[cls] = 1
            classification_labels.append(_classes)
            if len(cls) == 0:
                remove.append(ix)
                # print(f'{label} has no valid classification label')
                continue
        else:
            classification_labels.append(np.zeros(len(classes)))
            remove.append(ix)
            # print(f'{label} has no classification label')
    images = [i for ix, i in enumerate(images) if ix not in remove]
    classification_labels = [i for ix, i in enumerate(classification_labels) if ix not in remove]
    return images, classification_labels

class SCANEOClassificationDataModule(L.LightningDataModule):
    def __init__(self, 
            path, 
            classes = [], 
            batch_size=16, 
            num_workers=4, 
            pin_memory=True, 
            val_split=0.2, 
            train_trans=None, 
            val_trans=None
        ):
        super().__init__()
        self.path = path
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.train_trans = train_trans
        self.val_trans = val_trans
        self.val_split = val_split
        self.classes = classes
        self.num_classes = len(classes)

    def setup(self, stage=None):
        if len(self.classes) == 0:
            raise ValueError('a list of classes must be provided')
        labels = sorted(glob(f'{self.path}/*.geojson'))
        images = [i.replace('geojson', 'tif') for i in labels]
        images, classification_labels = get_classification_labels(images, labels, self.classes)
        train_image, val_image, train_label, val_label = train_test_split(images, classification_labels, test_size=self.val_split, random_state=42)
        print("Training on", len(train_image), "images")
        print("Validating on", len(val_image), "images")
        self.train_ds = ClassificationDataset(train_image, train_label, trans=self.train_trans)
        self.val_ds = ClassificationDataset(val_image, val_label, trans=self.val_trans)

    def train_dataloader(self, batch_size=None, shuffle=True):
        return DataLoader(
            self.train_ds, 
            batch_size=batch_size or self.batch_size, 
            num_workers=self.num_workers, 
            pin_memory=self.pin_memory,
            shuffle=shuffle
        )

    def val_dataloader(self, batch_size=None, shuffle=False):
        return DataLoader(
            self.val_ds, 
            batch_size=batch_size or self.batch_size, 
            num_workers=self.num_workers, 
            pin_memory=self.pin_memory,
            shuffle=shuffle
        )



import geopandas as gpd
import rasterio as rio
from rasterio.features import rasterize
from tqdm import tqdm
import json
import os
from glob import glob

def generate_masks_from_labels(path, verbose=True):

    labels_path = os.path.join(path, 'spai.json')
    if not os.path.exists(labels_path):
        raise FileNotFoundError(f"Labels file not found at {labels_path}")

    with open(labels_path, 'r') as f:
        spai_labels = json.load(f)
    colors = {label['name']: label['color'] for label in spai_labels['labels']}

    labels = sorted(glob(os.path.join(path, '*.geojson')))
    for _label in tqdm(labels):
        label = gpd.read_file(_label)
        seg = label[label['task'] == 'segmentation']
        if len(seg) == 0:
            if verbose:
                print(f"No segmentation labels found for {_label}")
            continue
        image = _label.replace('geojson', 'tif')
        if not os.path.exists(image):
            print(f"Image {image} not found for label {_label}")
            continue
        with rio.open(image) as ds:
            # target shape and transform
            out_shape = (ds.height, ds.width)
            transform = ds.transform
            # prepare (geometry, index) pairs
            # here we enumerate so indices start at 1; 0 will be background
            shapes = (
                (row.geometry, list(colors.keys()).index(row['label']) + 1)
                for _, row in seg.to_crs(ds.crs).iterrows()
            )
            # rasterize: background=0, polygons get their index
            mask = rasterize(
                shapes=shapes,
                out_shape=out_shape,
                transform=transform,
                fill=0,
                dtype='uint8',  # or 'uint16' if you have >255 polygons
            )
            # Save the mask with the same name as the image but with _mask suffix
            mask_path = image.replace('.tif', '_mask.tif')
            with rio.open(
                mask_path,
                'w',
                driver='GTiff',
                height=mask.shape[0],
                width=mask.shape[1],
                count=1,
                dtype=mask.dtype,
                crs=ds.crs,
                transform=transform,
            ) as dst:
                dst.write(mask, 1)


import pytest
try:
    from lib.eotdl import SHClient
except ImportError:
    from eotdl import SHClient


def test_sh_client():
    client = SHClient(sh_client_id='my_client_id',
    sh_client_secret='my_client_secret')

    assert client.config.sh_client_id is not None


def test_search_available_sentinel_data():
    pass


def test_request_bulk_data():
    pass


def test_request_data():
    pass



import pytest
import pystac
from datetime import datetime

from eotdl.curation.stac.extent import (
    get_dem_temporal_interval,
    get_unknow_temporal_interval,
    get_unknow_extent,
)


def test_get_dem_temporal_interval():
    expected_interval = [(datetime(2011, 1, 1, 0, 0), datetime(2015, 1, 7, 0, 0))]
    result_interval = get_dem_temporal_interval()
    assert result_interval.intervals == expected_interval


def test_get_unknow_temporal_interval():
    expected_interval = [(datetime(2000, 1, 1, 0, 0), datetime(2023, 12, 31, 0, 0))]
    result_interval = get_unknow_temporal_interval()
    assert result_interval.intervals == expected_interval


def test_get_unknow_extent():
    expected_spatial_extent = [[0, 0, 0, 0]]
    expected_temporal_extent = [
        (datetime(2000, 1, 1, 0, 0), datetime(2023, 12, 31, 0, 0))
    ]

    result_interval = get_unknow_extent()

    assert result_interval.spatial.bboxes == expected_spatial_extent
    assert result_interval.temporal.intervals == expected_temporal_extent



from pathlib import Path
import requests
import os

from ..repos import APIRepo


class FilesAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def ingest_file(
        self, file_path_or_bytes, file_name, dataset_or_model_id, user, endpoint, version=None
    ):
        url = self.url + f"{endpoint}/{dataset_or_model_id}"
        if version is not None:
            url += "?version=" + str(version)
        # get a presigned url to upload the file directly to the bucket
        reponse = requests.post(
            url,
            json={
                "file_name": file_name,
                # "file_size": files_size,
                # "checksum": checksum
            },
            headers=self.generate_headers(user),
        )
        data, error = self.format_response(reponse)
        if error:
            raise Exception(error)
        # ingest the file
        error = None
        try:
            presigned_url = data["presigned_url"]
            if isinstance(file_path_or_bytes, (str, bytes)):
                if isinstance(file_path_or_bytes, str):
                    # Handle file path
                    with open(file_path_or_bytes, 'rb') as f:
                        file_data = f.read()
                else:
                    # Handle bytes directly
                    file_data = file_path_or_bytes
                # Send file data to presigned URL
                response = requests.put(presigned_url, data=file_data)
                response.raise_for_status()
            else:
                raise TypeError("file_path_or_bytes must be either a file path string or bytes")
        except Exception as e:
            error = str(e)
        return data, error

    def stage_file(
        self,
        dataset_or_model_id,
        file_name,
        user,
        path,
        endpoint="datasets",
        progress=False,
    ):
        url = self.url + f"{endpoint}/{dataset_or_model_id}/stage/{file_name}"
        # if file_version is not None:
        #     url += "?version=" + str(file_version)
        return self.stage_file_url(url, path, user)

    def stage_file_url(
        self,
        url,
        path,
        user,
    ):
        if '/stage/' in url:  # asset is in EOTDL (can do better...)
            splitted = url.split("/stage/")
            file_name = splitted[-1]
            dataset_or_model_id = splitted[0].split("/")[-1]
            response = requests.get(url, headers=self.generate_headers(user))  # fixed typo
            #print(url)
            data, error = self.format_response(response)
            #print(data)
            if error:
                raise Exception(error)
            presigned_url = data["presigned_url"]
        else:
            splitted = url.split("//")
            file_name = splitted[-1]
            dataset_or_model_id = splitted[0].split("/")[-1]
            presigned_url = url

        file_path = f"{path}/{file_name}"
        for i in range(1, len(file_path.split("/")) - 1):
            os.makedirs("/".join(file_path.split("/")[: i + 1]), exist_ok=True)

        try:
            symbolic_link = False
            cache_base_path = os.getenv("EOTDL_CACHE_PATH", "")
            if cache_base_path:
                cache_path = Path(f"{cache_base_path}/{dataset_or_model_id}/{file_name}")
                #print(cache_path)
                if cache_path.exists():
                    symbolic_link = True
                    #print(f"Using cache: {cache_path}")

            if symbolic_link:
                if os.path.exists(file_path):
                    os.unlink(file_path)
                os.symlink(cache_path, file_path)
                #print(f"Symlinked {file_path} → {cache_path}")
            else:
                response = requests.get(presigned_url)
                response.raise_for_status()  # This will raise an HTTPError for 4XX and 5XX status codes
                with open(file_path, 'wb') as f:
                    f.write(response.content)
                #print(f"Downloaded to: {file_path}")
        except requests.exceptions.HTTPError as e:
            raise Exception(f"Failed to stage file: {str(e)}")
        except Exception as e:
            raise Exception(f"Unexpected error while staging file: {str(e)}")
        return file_path
    
    def generate_presigned_url(self, filename, dataset_or_model_id, user, endpoint="datasets"):
        url = f"{self.url}{endpoint}/{dataset_or_model_id}/stage/{filename}"
        reponse = requests.get(url, headers=self.generate_headers(user))
        data, error = self.format_response(reponse)
        if error:
            # print("ERROR generate_presigned_url", error)
            return None
        return data["presigned_url"]
    
    def generate_file_content_url(self, filename, dataset_or_model_id, user, endpoint="datasets"):
        url = f"{self.url}{endpoint}/{dataset_or_model_id}/raw/{filename}"
        return url


import lightning as L
import torch
import torch.nn as nn
from torchmetrics.classification import MultilabelAccuracy

from .model import MultiTaskModel
from ..segmentation.metrics import iou
    
# train a model for classification and segmentation at the same time

class MultiTaskModule(L.LightningModule):
    def __init__(self, num_seg_classes=4, num_cls_classes=4, encoder='resnet18', in_channels=3, pretrained=True, lr=1e-3):
        super().__init__()
        self.save_hyperparameters()
        self.model = MultiTaskModel(num_seg_classes, num_cls_classes, encoder, in_channels, pretrained)
        self.seg_loss = nn.BCEWithLogitsLoss()
        self.cls_loss = nn.BCEWithLogitsLoss()
        self.train_cls_acc = MultilabelAccuracy(num_labels=num_cls_classes)
        self.val_cls_acc = MultilabelAccuracy(num_labels=num_cls_classes)

    def forward(self, x):
        return self.model(x)
    
    def predict(self, x):
        self.eval()
        with torch.no_grad():
            seg_preds, cls_preds = self.model(x.to(self.device))
            return torch.sigmoid(seg_preds, dim=1), torch.argmax(cls_preds, dim=1)
    
    def training_step(self, batch, batch_idx):
        x, y_seg, y_cls = batch
        y_hat_seg, y_hat_cls = self(x)
        loss = self.seg_loss(y_hat_seg, y_seg) + self.cls_loss(y_hat_cls, y_cls)
        self.train_cls_acc(y_hat_cls, y_cls)
        metric = iou(y_hat_seg, y_seg)
        self.log('train_loss', loss, prog_bar=True)
        self.log('train_acc', self.train_cls_acc, prog_bar=True, logger=True)
        self.log('train_iou', metric, prog_bar=True, logger=True)
        return loss
    
    def validation_step(self, batch, batch_idx):
        x, y_seg, y_cls = batch
        y_hat_seg, y_hat_cls = self(x)
        loss = self.seg_loss(y_hat_seg, y_seg) + self.cls_loss(y_hat_cls, y_cls)
        self.val_cls_acc(y_hat_cls, y_cls)
        metric = iou(y_hat_seg, y_seg)
        self.log('val_loss', loss, prog_bar=True)
        self.log('val_acc', self.val_cls_acc, prog_bar=True, logger=True)
        self.log('val_iou', metric, prog_bar=True, logger=True)
        return loss

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.hparams.lr)


from ..repos import FEAPIRepo
from ..auth import with_auth
from .retrieve import retrieve_pipeline

@with_auth
def deactivate_pipeline(pipeline_name, user):
    pipeline = retrieve_pipeline(pipeline_name)
    repo = FEAPIRepo()
    data, error = repo.deactivate_pipeline(pipeline['id'], user)
    if error:
        raise Exception(error)
    return data



import torch
import torch.nn as nn
import math 
import torch.nn.functional as F

class ESRT(nn.Module):
	def __init__(self, upscale=2, in_channels=4):
		# super(ESRT, self).__init__()
		super().__init__()
		
		wn = lambda x: torch.nn.utils.weight_norm(x)
		
		n_feats = 32
		n_blocks = 1
		kernel_size = 3

		self.n_blocks = n_blocks
		
		# define head module
		modules_head = [conv(in_channels, n_feats, kernel_size)]
		
		# define body module
		modules_body = nn.ModuleList()
		for i in range(n_blocks):
			modules_body.append(Un(n_feats=n_feats, wn = wn))

		# define tail module
		modules_tail = [
			Upsampler(conv, upscale, n_feats, act=False),
			conv(n_feats, in_channels, kernel_size)
		]


		self.head = nn.Sequential(*modules_head)
		self.body = nn.Sequential(*modules_body)
		self.reduce = conv(n_blocks*n_feats, n_feats, kernel_size)
		self.tail = nn.Sequential(*modules_tail)
		self.up = nn.Sequential(
			Upsampler(conv,upscale,n_feats,act=False),
			conv(n_feats, in_channels, 3,1,1)
		)

	def forward(self, x):
		x = self.head(x)
		res2 = x
		body_out = []
		for i in range(self.n_blocks):
			x = self.body[i](x)
			body_out.append(x)
		res1 = torch.cat(body_out,1)
		res1 = self.reduce(res1)
		x = self.tail(res1)
		x = self.up(res2) + x
		return x
	
	def load_state_dict(self, state_dict, strict=False):
		own_state = self.state_dict()
		for name, param in state_dict.items():
			if name in own_state:
				if isinstance(param, nn.Parameter):
					param = param.data
				# If dimensions match, copy weights
				if own_state[name].size() == param.size():
					own_state[name].copy_(param)
				else:
					# If it's 'up' or 'tail', skip so those layers train from scratch
					if name.find('up') >= 0 or name.find('tail') >= 0:
						print(f'Skipping loading weights for "{name}" due to shape mismatch. Training from scratch...')
						continue
					else:
						raise RuntimeError(
							f'While copying the parameter named {name}, '
							f'whose dimensions in the model are {own_state[name].size()} and '
							f'whose dimensions in the checkpoint are {param.size()}.')
			elif strict:
				if name.find('tail') == -1 and name.find('up') == -1:
					raise KeyError(f'unexpected key "{name}" in state_dict')
		if strict:
			missing = set(own_state.keys()) - set(state_dict.keys())
			if len(missing) > 0:
				raise KeyError(f'missing keys in state_dict: "{missing}"')
	
def conv(in_channels, out_channels, kernel_size, bias=True, groups = 1):
	wn = lambda x: torch.nn.utils.weight_norm(x)
	return nn.Conv2d(
		in_channels, out_channels, kernel_size,
		padding=(kernel_size//2), bias=bias, groups = groups
	)

class Un(nn.Module):
	def __init__(self,n_feats, wn):
		super(Un, self).__init__()
		self.encoder1 = Updownblock(n_feats)
		self.encoder2 = Updownblock(n_feats)
		self.encoder3 = Updownblock(n_feats)
		self.reduce = conv(3*n_feats, n_feats, 3)
		self.weight2 = Scale(1)
		self.weight1 = Scale(1)
		self.attention = MLABlock(n_feat=n_feats, dim=288) 
		self.alise = conv(n_feats, n_feats, 3)

	def forward(self,x):
		# out = self.encoder3(self.encoder2(self.encoder1(x)))
		x1 = self.encoder1(x)
		x2 = self.encoder2(x1)
		x3 = self.encoder3(x2)
		out = x3
		b,c,h,w = x3.shape
		out = self.attention(self.reduce(torch.cat([x1,x2,x3],dim=1)))
		out = out.permute(0,2,1)
		out = reverse_patches(out, (h,w), (3,3), 1, 1)
		out = self.alise(out)

		return self.weight1(x) + self.weight2(out)
	
class Scale(nn.Module):

	def __init__(self, init_value=1e-3):
		super().__init__()
		self.scale = nn.Parameter(torch.FloatTensor([init_value]))

	def forward(self, input):
		return input * self.scale
	
class Upsampler(nn.Sequential):
	def __init__(self, conv, scale, n_feats, bn=False, act=False, bias=True):
		m = []
		
		# Handle scale factor as product of smaller factors
		if scale == 10:
			# First upscale by 2
			m.append(conv(n_feats, 4 * n_feats, 3, bias))
			m.append(nn.PixelShuffle(2))
			if bn: m.append(nn.BatchNorm2d(n_feats))
			if act == 'relu': m.append(nn.ReLU(True))
			elif act == 'prelu': m.append(nn.PReLU(n_feats))
			
			# Then upscale by 5
			m.append(conv(n_feats, 25 * n_feats, 3, bias))
			m.append(nn.PixelShuffle(5))
			if bn: m.append(nn.BatchNorm2d(n_feats))
			if act == 'relu': m.append(nn.ReLU(True))
			elif act == 'prelu': m.append(nn.PReLU(n_feats))
			
		elif (scale & (scale - 1)) == 0:    # Is scale = 2^n?
			for _ in range(int(math.log(scale, 2))):
				m.append(conv(n_feats, 4 * n_feats, 3, bias))
				m.append(nn.PixelShuffle(2))
				if bn: m.append(nn.BatchNorm2d(n_feats))
				if act == 'relu': m.append(nn.ReLU(True))
				elif act == 'prelu': m.append(nn.PReLU(n_feats))

		elif scale == 3:
			m.append(conv(n_feats, 9 * n_feats, 3, bias))
			m.append(nn.PixelShuffle(3))
			if bn: m.append(nn.BatchNorm2d(n_feats))
			if act == 'relu': m.append(nn.ReLU(True))
			elif act == 'prelu': m.append(nn.PReLU(n_feats))
			
		elif scale == 5:
			m.append(conv(n_feats, 25 * n_feats, 3, bias))
			m.append(nn.PixelShuffle(5))
			if bn: m.append(nn.BatchNorm2d(n_feats))
			if act == 'relu': m.append(nn.ReLU(True))
			elif act == 'prelu': m.append(nn.PReLU(n_feats))
			
		else:
			raise NotImplementedError(f"Scale factor {scale} is not supported")

		super(Upsampler, self).__init__(*m)

class Updownblock(nn.Module):
	def __init__(self, n_feats):
		super(Updownblock, self).__init__()
		self.encoder = one_module(n_feats)
		self.decoder_low = one_module(n_feats) #nn.Sequential(one_module(n_feats),
		#                     one_module(n_feats),
		#                     one_module(n_feats))
		self.decoder_high = one_module(n_feats)
		self.alise = one_module(n_feats)
		self.alise2 = BasicConv(2*n_feats, n_feats, 1,1,0) #one_module(n_feats)
		self.down = nn.AvgPool2d(kernel_size=2)
		self.att = CALayer(n_feats)

	def forward(self, x):
		x1 = self.encoder(x)
		x2 = self.down(x1)
		high = x1 - F.interpolate(x2, size = x.size()[-2:], mode='bilinear', align_corners=True)
		for i in range(5):
			x2 = self.decoder_low(x2)
		x3 = x2
		# x3 = self.decoder_low(x2)
		high1 = self.decoder_high(high)
		x4 = F.interpolate(x3, size = x.size()[-2:], mode='bilinear', align_corners=True)
		return self.alise(self.att(self.alise2(torch.cat([x4,high1],dim=1))))+ x
	
class one_module(nn.Module):
	def __init__(self, n_feats):
		super(one_module, self).__init__()
		self.layer1 = one_conv(n_feats, n_feats//2,3)
		self.layer2 = one_conv(n_feats, n_feats//2,3)
		# self.layer3 = one_conv(n_feats, n_feats//2,3)
		self.layer4 = BasicConv(n_feats, n_feats, 3,1,1)
		self.alise = BasicConv(2*n_feats, n_feats, 1,1,0)
		self.atten = CALayer(n_feats)
		self.weight1 = Scale(1)
		self.weight2 = Scale(1)
		self.weight3 = Scale(1)
		self.weight4 = Scale(1)
		self.weight5 = Scale(1)

	def forward(self, x):

		x1 = self.layer1(x)
		x2 = self.layer2(x1)
		# x3 = self.layer3(x2)
		# pdb.set_trace()
		x4 = self.layer4(self.atten(self.alise(torch.cat([self.weight2(x2),self.weight3(x1)],1))))
		return self.weight4(x)+self.weight5(x4)
	
class BasicConv(nn.Module):
	def __init__(self, in_planes, out_planes, kernel_size, stride=1, padding=1, dilation=1, groups=1, relu=True,
				 bn=False, bias=False, up_size=0,fan=False):
		super(BasicConv, self).__init__()
		wn = lambda x:torch.nn.utils.weight_norm(x)
		self.out_channels = out_planes
		self.in_channels = in_planes
		if fan:
			self.conv = nn.ConvTranspose2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding,
							  dilation=dilation, groups=groups, bias=bias)
		else:
			self.conv = nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding,
							  dilation=dilation, groups=groups, bias=bias)
		self.bn = nn.BatchNorm2d(out_planes, eps=1e-5, momentum=0.01, affine=True) if bn else None
		self.relu = nn.ReLU(inplace=True) if relu else None
		self.up_size = up_size
		self.up_sample = nn.Upsample(size=(up_size, up_size), mode='bilinear') if up_size != 0 else None
		
	def forward(self, x):
		x = self.conv(x)
		if self.bn is not None:
			x = self.bn(x)
		if self.relu is not None:
			x = self.relu(x)
		if self.up_size > 0:
			x = self.up_sample(x)
		return x
	
class one_conv(nn.Module):
	def __init__(self,inchanels,growth_rate,kernel_size = 3, relu = True):
		super(one_conv,self).__init__()
		wn = lambda x:torch.nn.utils.weight_norm(x)
		self.conv = nn.Conv2d(inchanels,growth_rate,kernel_size=kernel_size,padding = kernel_size>>1,stride= 1)
		self.flag = relu
		self.conv1 = nn.Conv2d(growth_rate,inchanels,kernel_size=kernel_size,padding = kernel_size>>1,stride= 1)
		if relu:
			self.relu = nn.PReLU(growth_rate)
		self.weight1 = Scale(1)
		self.weight2 = Scale(1)
	def forward(self,x):
		if self.flag == False:
			output = self.weight1(x) + self.weight2(self.conv1(self.conv(x)))
		else:
			output = self.weight1(x) + self.weight2(self.conv1(self.relu(self.conv(x))))
		return output#torch.cat((x,output),1)
	
## Channel Attention (CA) Layer
class CALayer(nn.Module):
	def __init__(self, channel, reduction=16):
		super(CALayer, self).__init__()
		# global average pooling: feature --> point
		self.avg_pool = nn.AdaptiveAvgPool2d(1)
		# feature channel downscale and upscale --> channel weight
		self.conv_du = nn.Sequential(
				nn.Conv2d(channel, channel // reduction, 1, padding=0, bias=True),
				nn.ReLU(inplace=True),
				nn.Conv2d(channel // reduction, channel, 1, padding=0, bias=True),
				nn.Sigmoid()
		)

	def forward(self, x):
		y = self.avg_pool(x)
		y = self.conv_du(y)
		return x * y
	
class MLABlock(nn.Module):
	def __init__(
		self, n_feat = 64,dim=768, num_heads=8, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
				 drop_path=0., act_layer=nn.ReLU, norm_layer=nn.LayerNorm):
		super(MLABlock, self).__init__()
		self.dim = dim
		self.atten = EffAttention(self.dim, num_heads=8, qkv_bias=False, qk_scale=None, \
							 attn_drop=0., proj_drop=0.)
		self.norm1 = nn.LayerNorm(self.dim)
		# self.posi = PositionEmbeddingLearned(n_feat)
		self.mlp = Mlp(in_features=dim, hidden_features=dim//4, act_layer=act_layer, drop=drop)
		self.norm2 = nn.LayerNorm(self.dim)
	def forward(self, x):
		# pdb.set_trace()
		B = x.shape[0]
		# posi = self.posi(x)
		# x = posi + x
		# x = self.patch_embed(x) # 16*36*768
		# print(x.shape)
		x = extract_image_patches(x, ksizes=[3, 3],
									  strides=[1,1],
									  rates=[1, 1],
									  padding='same')#   16*2304*576
		x = x.permute(0,2,1)

		x = x + self.atten(self.norm1(x))
		x = x + self.mlp(self.norm2(x))#self.drop_path(self.mlp(self.norm2(x)))
		return x
	
class EffAttention(nn.Module):
	def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
		super().__init__()
		self.num_heads = num_heads
		head_dim = dim // num_heads
		# NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
		self.scale = qk_scale or head_dim ** -0.5

		self.reduce = nn.Linear(dim, dim//2, bias=qkv_bias)
		self.qkv = nn.Linear(dim//2, dim//2 * 3, bias=qkv_bias)
		self.proj = nn.Linear(dim//2, dim)
		self.attn_drop = nn.Dropout(attn_drop)
		# print('scale', self.scale)
		# print(dim)
		# print(dim//num_heads)
		# self.proj = nn.Linear(dim, dim)
		# self.proj_drop = nn.Dropout(proj_drop)

	def forward(self, x):
		x = self.reduce(x)
		B, N, C = x.shape
		# pdb.set_trace()
		qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
		# q = x.reshape(B, N, 1, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
		# k = x.reshape(B, N, 1, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
		# v = x.reshape(B, N, 1, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
		# qkv: 3*16*8*37*96
		q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)
		# pdb.set_trace()
		
		q_all = torch.split(q, math.ceil(N//4), dim=-2)
		k_all = torch.split(k, math.ceil(N//4), dim=-2)
		v_all = torch.split(v, math.ceil(N//4), dim=-2)        

		output = []
		for q,k,v in zip(q_all, k_all, v_all):
			attn = (q @ k.transpose(-2, -1)) * self.scale   #16*8*37*37
			attn = attn.softmax(dim=-1)
			attn = self.attn_drop(attn)
			
			trans_x = (attn @ v).transpose(1, 2)#.reshape(B, N, C)

			output.append(trans_x)
		# pdb.set_trace()
		# attn = torch.cat(att, dim=-2)
		# x = (attn @ v).transpose(1, 2).reshape(B, N, C) #16*37*768
		x = torch.cat(output,dim=1)
		x = x.reshape(B,N,C)
		# pdb.set_trace()
		x = self.proj(x)
		# x = self.proj_drop(x)
		return x
	
class Mlp(nn.Module):
	def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.ReLU, drop=0.):
		super().__init__()
		out_features = out_features or in_features
		hidden_features = hidden_features or in_features//4
		self.fc1 = nn.Linear(in_features, hidden_features)
		self.act = act_layer()
		self.fc2 = nn.Linear(hidden_features, out_features)
		self.drop = nn.Dropout(drop)

	def forward(self, x):
		x = self.fc1(x)
		x = self.act(x)
		x = self.drop(x)
		x = self.fc2(x)
		x = self.drop(x)
		return x
	
def extract_image_patches(images, ksizes, strides, rates, padding='same'):
	"""
	Extract patches from images and put them in the C output dimension.
	:param padding:
	:param images: [batch, channels, in_rows, in_cols]. A 4-D Tensor with shape
	:param ksizes: [ksize_rows, ksize_cols]. The size of the sliding window for
	 each dimension of images
	:param strides: [stride_rows, stride_cols]
	:param rates: [dilation_rows, dilation_cols]
	:return: A Tensor
	"""
	assert len(images.size()) == 4
	assert padding in ['same', 'valid']
	batch_size, channel, height, width = images.size()
	
	if padding == 'same':
		images = same_padding(images, ksizes, strides, rates)
	elif padding == 'valid':
		pass
	else:
		raise NotImplementedError('Unsupported padding type: {}.\
				Only "same" or "valid" are supported.'.format(padding))

	unfold = torch.nn.Unfold(kernel_size=ksizes,
							 dilation=rates,
							 padding=0,
							 stride=strides)
	patches = unfold(images)
	return patches  # [N, C*k*k, L], L is the total number of such blocks

def reverse_patches(images, out_size, ksizes, strides, padding):
	"""
	Extract patches from images and put them in the C output dimension.
	:param padding:
	:param images: [batch, channels, in_rows, in_cols]. A 4-D Tensor with shape
	:param ksizes: [ksize_rows, ksize_cols]. The size of the sliding window for
	 each dimension of images
	:param strides: [stride_rows, stride_cols]
	:param rates: [dilation_rows, dilation_cols]
	:return: A Tensor
	"""
	unfold = torch.nn.Fold(output_size = out_size, 
							kernel_size=ksizes, 
							dilation=1, 
							padding=padding, 
							stride=strides)
	patches = unfold(images)
	return patches  # [N, C*k*k, L], L is the total number of such blocks

def same_padding(images, ksizes, strides, rates):
	assert len(images.size()) == 4
	batch_size, channel, rows, cols = images.size()
	out_rows = (rows + strides[0] - 1) // strides[0]
	out_cols = (cols + strides[1] - 1) // strides[1]
	effective_k_row = (ksizes[0] - 1) * rates[0] + 1
	effective_k_col = (ksizes[1] - 1) * rates[1] + 1
	padding_rows = max(0, (out_rows-1)*strides[0]+effective_k_row-rows)
	padding_cols = max(0, (out_cols-1)*strides[1]+effective_k_col-cols)
	# Pad the input
	padding_top = int(padding_rows / 2.)
	padding_left = int(padding_cols / 2.)
	padding_bottom = padding_rows - padding_top
	padding_right = padding_cols - padding_left
	paddings = (padding_left, padding_right, padding_top, padding_bottom)
	images = torch.nn.ZeroPad2d(paddings)(images)
	return images



from ..repos import ModelsAPIRepo
from ..auth import with_auth
from .retrieve import retrieve_model
def update_model(model_id, metadata, content, user):
    repo = ModelsAPIRepo()
    data, error = repo.update_model(
        model_id,
        metadata.authors,
        metadata.source,
        metadata.license,
        metadata.thumbnail,
        content,
        user,
    )
    if error:
        raise Exception(error)
    return data

@with_auth
def deactivate_model(model_name, user):
    model = retrieve_model(model_name)
    repo = ModelsAPIRepo()
    data, error = repo.deactivate_model(model['id'], user)
    if error:
        raise Exception(error)
    return data



import json
from bson import ObjectId
from minio import Minio
from pymongo import MongoClient
import pytest
import boto3


@pytest.fixture
def setup_mongo():
    client = MongoClient("mongodb://localhost:27017/")
    db = client["eotdl-test"]

    client.drop_database("eotdl-test")

    db = client["eotdl-test"]
    tiers_collection = db["tiers"]

    with open("eotdl/tests/load/eotdl.tiers.json", "r") as file:
        json_data = json.load(file)

    for item in json_data:
        if "_id" in item:
            item["_id"] = ObjectId(item["_id"])

    tiers_collection.insert_many(json_data)

    yield tiers_collection

    client.drop_database("eotdl-test")


@pytest.fixture
def setup_minio():
    minio_client = Minio(
        "localhost:9000",
        access_key="eotdl",
        secret_key="12345678",
        secure=False, 
    )

    bucket_name = "eotdl-test"

    if not minio_client.bucket_exists(bucket_name):
        minio_client.make_bucket(bucket_name)

    objects = minio_client.list_objects(bucket_name, recursive=True)
    for obj in objects:
        minio_client.remove_object(bucket_name, obj.object_name)

    yield minio_client

    objects = minio_client.list_objects(bucket_name, recursive=True)
    for obj in objects:
        minio_client.remove_object(bucket_name, obj.object_name)



import torch


def iou(pr, gt, th=0.5, eps=1e-7):
    pr = torch.sigmoid(pr) > th
    intersection = torch.sum(gt * pr, axis=(-2, -1))
    union = (
        torch.sum(gt, axis=(-2, -1)) + torch.sum(pr, axis=(-2, -1)) - intersection + eps
    )
    ious = (intersection + eps) / union
    return torch.mean(ious)


import pytest
from unittest import mock
from unittest.mock import patch

from eotdl.src.usecases.datasets.DownloadDataset import DownloadDataset


@pytest.fixture
def dataset():
    return {
        "name": "test",
        "id": "test-id",
        "files": [{"name": "test-file", "size": 123, "checksum": "123"}],
    }


@pytest.fixture
def user():
    return {"id_token": "test token"}



def _find_matches(args):
    from .matches import find_sentinel_matches_by_centroid
    ix, centroid, date, time_buffer, width, height, collection_id = args
    return (ix, find_sentinel_matches_by_centroid(centroid, date, time_buffer, width, height, collection_id))

def _download_sentinel_imagery(args):
    from .download import download_sentinel_imagery
    dir, date, bb, col, name = args
    download_sentinel_imagery(dir, date, bb, col, name)


import os
import shutil
import pytest

from pystac import Catalog

EXAMPLE_DATA_DIR = "tests/unit/curation/stac/test_data"


@pytest.fixture
def tmp_stac_catalog():
    original_catalog_path = os.path.join(EXAMPLE_DATA_DIR, "jaca_dataset_stac")

    os.makedirs("tmp", exist_ok=True)
    tmp_catalog_path = os.path.join("tmp", "copied_catalog")
    tmp_catalog = Catalog.from_file(os.path.join(original_catalog_path, "catalog.json"))
    tmp_catalog.normalize_and_save(tmp_catalog_path)

    copied_catalog = Catalog.from_file(os.path.join(tmp_catalog_path, "catalog.json"))
    yield copied_catalog

    shutil.rmtree("tmp")


@pytest.fixture
def tmp_stac_catalog_labels():
    original_catalog_path = os.path.join(EXAMPLE_DATA_DIR, "jaca_dataset_stac_labels")

    os.makedirs("tmp", exist_ok=True)
    tmp_catalog_path = os.path.join("tmp", "copied_catalog")
    tmp_catalog = Catalog.from_file(os.path.join(original_catalog_path, "catalog.json"))
    tmp_catalog.normalize_and_save(tmp_catalog_path)

    copied_catalog = Catalog.from_file(os.path.join(tmp_catalog_path, "catalog.json"))
    yield copied_catalog

    shutil.rmtree("tmp")


@pytest.fixture
def tmp_stac_collection():
    original_collection_path = os.path.join(
        EXAMPLE_DATA_DIR, "jaca_dataset_stac", "collection.json"
    )

    os.makedirs("tmp", exist_ok=True)
    tmp_collection_path = os.path.join("tmp", "copied_collection")
    shutil.copy(original_collection_path, tmp_collection_path)

    copied_collection = Catalog.from_file(
        os.path.join(tmp_collection_path, "collection.json")
    )
    yield copied_collection

    shutil.rmtree("tmp")


@pytest.fixture
def sentinel_2():
    os.makedirs("tmp", exist_ok=True)

    yield os.path.join(EXAMPLE_DATA_DIR, 'sentinel_2')

    shutil.rmtree("tmp")


@pytest.fixture
def tmp_sentinel_2():
    original_dataset_path = os.path.join(EXAMPLE_DATA_DIR, "sentinel_2")

    os.makedirs("tmp", exist_ok=True)
    shutil.copytree(original_dataset_path, "tmp/sentinel_2")

    yield "tmp/sentinel_2"

    shutil.rmtree("tmp")



from .retrieve import retrieve_datasets, retrieve_dataset, retrieve_dataset_files, retrieve_private_datasets
from .ingest import ingest_dataset, ingest_virtual_dataset, ingest_dataset_catalog
from .stage import stage_dataset, stage_dataset_file
from .update import deactivate_dataset


from ..repos import ModelsAPIRepo, FilesAPIRepo


def retrieve_models(name=None, limit=None):
    api_repo = ModelsAPIRepo()
    data, error = api_repo.retrieve_models(name, limit)
    if data and not error:
        models = [d["name"] for d in data] if data else []
        return models
    return []


def retrieve_model(name):
    repo = ModelsAPIRepo()
    data, error = repo.retrieve_model(name)
    if error:
        raise Exception(error)
    return data


def retrieve_model_files(model_id, version):
    repo = FilesAPIRepo()
    data, error = repo.retrieve_files(model_id, "models", version)
    if error:
        raise Exception(error)
    return data



import pytest
from unittest import mock

from eotdl.src.usecases.datasets.RetrieveDatasets import RetrieveDatasets
from eotdl.src.usecases.datasets.RetrieveDataset import RetrieveDataset


def test_retrieve_datasets():
    repo = mock.Mock()
    retrieve = RetrieveDatasets(repo)
    inputs = RetrieveDatasets.Inputs()
    repo.retrieve_datasets.return_value = [
        {"name": "test1", "files": [{"name": "test11"}]},
        {"name": "test2", "files": [{"name": "test21"}, {"name": "test22"}]},
    ]
    outputs = retrieve(inputs)
    repo.retrieve_datasets.assert_called_once()
    assert outputs.datasets == {
        "test1": ["test11"],
        "test2": ["test21", "test22"],
    }


def test_retrieve_dataset():
    repo = mock.Mock()
    retrieve = RetrieveDataset(repo)
    inputs = RetrieveDataset.Inputs(name="test")
    repo.retrieve_dataset.return_value = {"name": "test"}, None
    outputs = retrieve(inputs)
    assert outputs.dataset == {"name": "test"}
    repo.retrieve_dataset.assert_called_once_with("test")


def test_retrieve_dataset_fails():
    repo = mock.Mock()
    retrieve = RetrieveDataset(repo)
    inputs = RetrieveDataset.Inputs(name="test")
    repo.retrieve_dataset.return_value = None, "error"
    with pytest.raises(Exception):
        retrieve(inputs)
    repo.retrieve_dataset.assert_called_once_with("test")


# def test_download_dataset():
#     repo = mock.Mock()
#     download = DownloadDataset(repo)
#     inputs = DownloadDataset.Inputs(
#         dataset="test", path="test path", user={"id_token": "test token"}
#     )
#     repo.download_dataset.return_value = "dst path"
#     outputs = download(inputs)
#     assert outputs.dst_path == "dst path"
#     repo.download_dataset.assert_called_once_with("test", "test token", "test path")


# def test_ingest_dataset_fails_if_not_zip():
#     repo = mock.Mock()
#     logger = mock.Mock()
#     ingest = IngestDataset(repo, logger)
#     inputs = IngestDataset.Inputs(
#         name="test",
#         description="test description",
#         path="test path",
#         user={"id_token": "test token"},
#     )
#     with pytest.raises(Exception):
#         ingest(inputs)


# def test_ingest_dataset():
#     repo = mock.Mock()
#     repo.ingest_dataset.return_value = {"name": "test"}, None
#     logger = mock.Mock()
#     ingest = IngestDataset(repo, logger)
#     inputs = IngestDataset.Inputs(
#         name="test",
#         description="test description",
#         path="test_path.zip",
#         user={"id_token": "test token"},
#     )
#     outputs = ingest(inputs)
#     repo.ingest_dataset.assert_called_once_with(
#         "test", "test description", "test_path.zip", "test token"
#     )
#     assert outputs.dataset == {"name": "test"}


# def test_ingest_large_dataset_fails_if_not_zip():
#     repo = mock.Mock()
#     logger = mock.Mock()
#     ingest = IngestLargeDataset(repo, logger)
#     inputs = IngestLargeDataset.Inputs(
#         name="test",
#         description="test description",
#         path="test path",
#         user={"id_token": "test token"},
#     )
#     with pytest.raises(Exception):
#         ingest(inputs)


# def test_ingest_large_dataset():
#     repo = mock.Mock()
#     repo.ingest_large_dataset.return_value = {"name": "test"}, None
#     logger = mock.Mock()
#     ingest = IngestLargeDataset(repo, logger)
#     inputs = IngestLargeDataset.Inputs(
#         name="test",
#         description="test description",
#         path="test_path.zip",
#         user={"id_token": "test token"},
#     )
#     outputs = ingest(inputs)
#     repo.ingest_large_dataset.assert_called_once_with(
#         "test", "test description", "test_path.zip", "test token"
#     )
#     assert outputs.dataset == {"name": "test"}



from eotdl.tools.tools import *


def test_extract_image_id_in_folder():
    assert extract_image_id_in_folder("path/to/image/20210101T000000.tif", 3) == "20210101T000000.tif"
    assert extract_image_id_in_folder("path/to/image/20210101T000000.tif", 2) == "image"



import pytest
from datetime import datetime

from eotdl.tools.time_utils import *


@pytest.mark.parametrize("input_time_interval, expected_output", [
    (["2022-01-01", "2022-01-10"], True),
    (["2021-05-15", "2021-05-16"], True),
    ([], False),
    (["2022-01-01"], False),
    (["2022-01-01", "2022-01-10", "2022-01-11"], False),
    ([2022, 1], False),
    ("2022-01-01,2022-01-10", False),
    (["08:00", "18:00"], False),
    (["2022-01-32", "2022-01-10"], False),
])
def test_is_time_interval(input_time_interval, expected_output):
    assert is_time_interval(input_time_interval) == expected_output


@pytest.mark.parametrize("start_date, end_date, n_chunks, expected_output", [
    (
        datetime(2022, 1, 1),
        datetime(2022, 1, 3),
        3,
        [('2022-01-01', '2022-01-01'), ('2022-01-01', '2022-01-02')]
    ),
    (
        datetime(2022, 1, 1),
        datetime(2022, 1, 6),
        3,
        [('2022-01-01', '2022-01-02'), ('2022-01-02', '2022-01-04')]
    )
])
def test_create_time_slots(start_date, end_date, n_chunks, expected_output):
    assert create_time_slots(start_date, end_date, n_chunks) == expected_output


@pytest.mark.parametrize("input_date, expected_output", [
    ("2022-01-01", True),
    ("2000-02-29", True),
    ("2022-01-32", False),
    ("2022-13-01", False),
    ("2022-02-30", False),
    ("2001-02-29", False),
    ("22-01-01", False),
    ("", False),
    ("NotADate", False),
])
def test_is_valid_date(input_date, expected_output):
    assert is_valid_date(input_date) == expected_output


@pytest.mark.parametrize(
    "from_date, to_date, expected",
    [
        ("2019-12-31T00:00:00Z", "2020-01-02T23:59:59Z", "2020-01-01"),
        (datetime(2019, 12, 31, 0, 0), datetime(2020, 1, 2, 23, 59, 59), "2020-01-01"),
        ("2020-01-01T00:00:00Z", "2020-01-03T23:59:59Z", "2020-01-02"),
        (datetime(2020, 1, 1, 0, 0), datetime(2020, 1, 3, 23, 59, 59), "2020-01-02"),
    ]
)
def test_get_day_between(from_date, to_date, expected):
    assert get_day_between(from_date, to_date) == expected


def test_expand_time_interval():
    pass


def test_prepare_time_interval():
    # Test para input en formato string
    assert prepare_time_interval("2022-01-01") == ("2021-12-31", "2022-01-02")
    
    # Test para input en formato datetime
    test_date = datetime.strptime("2022-01-01", "%Y-%m-%d")
    assert prepare_time_interval(test_date) == ("2021-12-31", "2022-01-02")
    
    # Test para input en formato de tuple válido
    assert prepare_time_interval(("2022-01-01", "2022-01-02")) == ("2022-01-01", "2022-01-02")
    
    # Test para input en formato de tuple inválido
    with pytest.raises(ValueError, match=r".*range of two dates.*"):
        prepare_time_interval(("2022-01-01",))
        
    # Test para input no válido
    with pytest.raises(ValueError, match=r".*string with format YYYY-MM-DD or a datetime object.*"):
        prepare_time_interval(1234)


@pytest.mark.parametrize(
    "dt, expected",
    [
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
        ("2021-01-01", datetime(2021, 1, 1, 0, 0)),
    ],
)
def test_format_time_acquired(dt, expected):
    assert format_time_acquired(dt) == expected
